/*
 * Copyright (c) 2007-2015 Freescale Semiconductor, Inc.
 * Copyright 2018-2019 NXP
 *
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * FreeMASTER Communication Driver - Example Application
 */

  /*******************************************************************************
   * Includes
   ******************************************************************************/
  #include <math.h>

  #include "fsl_debug_console.h"
  #include "board.h"
  #include "pin_mux.h"
  #include "fsl_common.h"
  #include "fsl_clock.h"

  #include "fsl_adc16.h"
  #include "fsl_tpm.h"
  #include "fsl_gpio.h"

  #include "freemaster.h"

  /*******************************************************************************
   * ADC Definitions and variables
   ******************************************************************************/
  #define DEMO_ADC16_BASE ADC0
  #define DEMO_ADC16_CHANNEL_GROUP 0U

  #define DEMO_ADC16_USER_CHANNEL0 1U   /* PTE16, J4-2  on FRDM-KL27Z. */
  #define DEMO_ADC16_USER_CHANNEL1 14U  /* PTC00, J4-4  on FRDM-KL27Z. */
  #define DEMO_ADC16_USER_CHANNEL2 0U   /* PTE20, J4-6  on FRDM-KL27Z. */
  #define DEMO_ADC16_USER_CHANNEL3 4U   /* PTE21, J4-8  on FRDM-KL27Z. */
  #define DEMO_ADC16_USER_CHANNEL4 9U   /* PTB01, J4-10 on FRDM-KL27Z. */

  #define VCC_mV 3300.0
  #define ADC_range 4095.0
  #define R_Resistor 4700.0
  #define R_Ref_NTC 10000.0
  #define A1 3.354016E-03
  #define B1 2.569850E-04
  #define C1 2.620131E-06
  #define D1 6.383091E-08

  uint16_t ADC_Value = 0;
  double U_Value = 0.0;
  double R_NTC = 0.0;
  double T_Value = 0.0;
  double C_Value = 0.0;
  double I_Value = 0.0;

  adc16_channel_config_t adc16ChannelConfigStruct;

  /*******************************************************************************
   * TPM Definitions and variables
   ******************************************************************************/
  /* define instance */
  #define BOARD_TPM TPM2
  /* Interrupt number and interrupt handler for the TPM instance used */
  #define BOARD_TPM_IRQ_NUM TPM2_IRQn
  #define BOARD_TPM_HANDLER TPM2_IRQHandler
  /* Get source clock for TPM driver */
  #define TPM_SOURCE_CLOCK (CLOCK_GetFreq(kCLOCK_McgIrc48MClk)/4)

  uint8_t interrupts = 0;

  #define PORTA_IRQ_NUM PORTA_IRQn
  #define PORTA_IRQ_HANDLER PORTA_IRQHandler

  /*******************************************************************************
   * Variables of measured and limit values
   ******************************************************************************/
  double temperature1 = 0;
  double temperature2 = 0;
  double temperature3 = 0;
  double temperature4 = 0;
  double current = 0;

  double correctionValue = 0; //value for correction of current measurement

  //default limit values
  double limit_temperature1 = 80.0;
  double limit_temperature2 = 80.0;
  double limit_temperature3 = 80.0;
  double limit_temperature4 = 80.0;
  double limit_current = 5.0;

  /*******************************************************************************
  * Variables of measured and limit values
  ******************************************************************************/
  uint16_t disconnectionSrc = 0;

  uint16_t relayState = 0; //1 - RC, 0 - RO
  uint16_t oldRelayState = 0;
  uint16_t resetState = 0;

  bool buzzer = false;
  uint16_t buzzerTimer = 0;

  bool interruptDisabled = false;
  uint16_t interruptTimer = 0;

  uint16_t testValue; //value for checkConnection();

  const uint16_t applicationID = 40284;

  /*******************************************************************************
   * Prototypes
   ******************************************************************************/
  static void init_ADC16(void);
  static void init_TPM(void);
  void relayDisconnect(uint16_t source);
  void relayConnect(void);
  void checkLimits(void);
  double temperatureMeasurement(adc16_channel_config_t *adc16ChannelConfigStruct);
  void currentMeasurement(void);
  void currentCorrection(void);
  void setBuzzer(bool state);

  /*******************************************************************************
   * Code
   ******************************************************************************/

int main(void)
{
    /* Board initialization */
    BOARD_InitPins();
    BOARD_BootClockRUN();
    BOARD_InitDebugConsole();

    /* Periodic timer initialization */
    init_TPM();

    /* ADC module initialization */
    init_ADC16();

    /* FreeMASTER initialization */
    FMSTR_Init();

    /* Interrupt enable */
    EnableIRQ(BOARD_TPM_IRQ_NUM);
    EnableIRQ(PORTA_IRQ_NUM);

    relayDisconnect(0);
    currentCorrection();

    resetState = 1;
    testValue = 0;

    while (1)
    {
        FMSTR_Poll();

        checkLimits();

        if(relayState != oldRelayState)
        {
          if((relayState == 1) && ((temperature1 < limit_temperature1) && (temperature2 < limit_temperature2) && (temperature3 < limit_temperature3) && (temperature4 < limit_temperature4) && (current < limit_current)))
            relayConnect(); //maual connection from FreeMASTER GUI
          else if(relayState == 0)
            relayDisconnect(6); //maual disconnection
          else
          {
            relayState = 0;
          }
          oldRelayState = relayState;
        }
    }
}


/*!
 * brief Initialize ADC module with default configuration.
 *
 */
  static void init_ADC16(void)
  {
    adc16_config_t adc16ConfigStruct;

    ADC16_GetDefaultConfig(&adc16ConfigStruct);
    ADC16_Init(DEMO_ADC16_BASE, &adc16ConfigStruct);
    ADC16_EnableHardwareTrigger(DEMO_ADC16_BASE, false);
    ADC16_DoAutoCalibration(DEMO_ADC16_BASE);

    adc16ChannelConfigStruct.enableInterruptOnConversionCompleted = false;
    adc16ChannelConfigStruct.enableDifferentialConversion = false;
  }

/*!
 * brief Initialize TPM module.
 *
 * Initializes TMP and sets divider to 128.
 */
  static void init_TPM(void)
  {
    tpm_config_t tpmInfo;

    /* Select the clock source for the TPM counter as MCGPLLCLK */
    CLOCK_SetTpmClock(1U);

    TPM_GetDefaultConfig(&tpmInfo);

    #ifndef TPM_PRESCALER
    #define TPM_PRESCALER kTPM_Prescale_Divide_128
    #endif

    /* TPM clock divide by TPM_PRESCALER */
    tpmInfo.prescale = TPM_PRESCALER;

    /* Initialize TPM module */
    TPM_Init(BOARD_TPM, &tpmInfo);

    /*
     * Set timer period.
     */
    TPM_SetTimerPeriod(BOARD_TPM, USEC_TO_COUNT(1000U, TPM_SOURCE_CLOCK));
    TPM_EnableInterrupts(BOARD_TPM, kTPM_TimeOverflowInterruptEnable);
    TPM_StartTimer(BOARD_TPM, kTPM_SystemClock);
  }

/*!
 * brief Checks if the limit values do not exceed the maximum permissible values.
 */
   void checkLimits()
   {
     if(limit_temperature1 > 110.0)
       limit_temperature1 = 110.0;

     if(limit_temperature2 > 110.0)
       limit_temperature2 = 110.0;

     if(limit_temperature3 > 110.0)
       limit_temperature3 = 110.0;

     if(limit_temperature4 > 110.0)
       limit_temperature4 = 110.0;

     if(limit_current > 5.0)
       limit_current = 5.0;
   }

/*!
 * brief Disconnects the relay.
 *
 * Disconnects the relay, sets LEDs and turns on the buzzer.
 *
 * param source Indicates which source caused disconnection of relay.
 */
   void relayDisconnect(uint16_t source)
   {
      GPIO_PortClear(BOARD_INITPINS_Relay_GPIO, 1u << BOARD_INITPINS_Relay_PIN);        //disconnection of relay

      GPIO_PortClear(BOARD_INITPINS_GREEN_GPIO, 1u << BOARD_INITPINS_GREEN_PIN);          //GREEN LED OFF
      GPIO_PortSet(BOARD_INITPINS_RED_GPIO, 1u << BOARD_INITPINS_RED_PIN);            //RED LED ON

      setBuzzer(true);

      disconnectionSrc = source;

      relayState = 0;
      oldRelayState = 0;
   }

/*!
 * brief Connects the relay.
 *
 * Connects the relay and sets LEDs.
 */
  void relayConnect()
  {
      GPIO_PortSet(BOARD_INITPINS_Relay_GPIO, 1u << BOARD_INITPINS_Relay_PIN);           //connection of relay

      GPIO_PortSet(BOARD_INITPINS_GREEN_GPIO, 1u << BOARD_INITPINS_GREEN_PIN);          //GREEN LED OFF
      GPIO_PortClear(BOARD_INITPINS_RED_GPIO, 1u << BOARD_INITPINS_RED_PIN);            //RED LED ON

      disconnectionSrc = 0; //no source of disconnection

      relayState = 1;
      oldRelayState = 1;
  }

/*!
 * brief Temperature measurement.
 *
 * Performs voltage measurements at the appropriate input and converts the measured value to temperature in degrees Celsius.
 *
 * param *adc16ChannelConfigStruct Pointer to configuration structure of ADC module
 * return C_Value Temperature in degrees Celsius.
 */
  double temperatureMeasurement(adc16_channel_config_t *adc16ChannelConfigStruct)
  {
      ADC16_SetChannelConfig(DEMO_ADC16_BASE, DEMO_ADC16_CHANNEL_GROUP, adc16ChannelConfigStruct);
      while (0U == (kADC16_ChannelConversionDoneFlag & ADC16_GetChannelStatusFlags(DEMO_ADC16_BASE, DEMO_ADC16_CHANNEL_GROUP))) {}

      double Logarithm = 0.0;

      ADC_Value = ADC16_GetChannelConversionValue(DEMO_ADC16_BASE, DEMO_ADC16_CHANNEL_GROUP);   //Value measured from ADC0
      U_Value = ((ADC_Value/ADC_range) * VCC_mV);   // U_Value from ADC0 in mV
      R_NTC = (-(R_Resistor * U_Value)/(U_Value - VCC_mV));

      Logarithm = log(R_NTC/R_Ref_NTC);
      T_Value = (1/(A1 + B1 * Logarithm + C1 * pow(Logarithm, 2) + D1 * pow(Logarithm, 3)));
      C_Value = T_Value - 273.15;

      return C_Value;
  }

/*!
 * brief Current measurement.
 *
 * Performs voltage measurements at the appropriate input and converts the measured value to current in amperes. After that it checks if measured value exceeded limit value.
 *
 */
  void currentMeasurement()
  {
      adc16ChannelConfigStruct.channelNumber = DEMO_ADC16_USER_CHANNEL4;      //ADC4 (current (5))

      ADC16_SetChannelConfig(DEMO_ADC16_BASE, DEMO_ADC16_CHANNEL_GROUP, &adc16ChannelConfigStruct);
      while (0U == (kADC16_ChannelConversionDoneFlag & ADC16_GetChannelStatusFlags(DEMO_ADC16_BASE, DEMO_ADC16_CHANNEL_GROUP))) {}

      ADC_Value = ADC16_GetChannelConversionValue(DEMO_ADC16_BASE, DEMO_ADC16_CHANNEL_GROUP);   //Value measured from ADC0
      U_Value = ((ADC_Value/ADC_range) * VCC_mV);   // U_Value from ADC0 in mV

      U_Value = U_Value - correctionValue; //correctionValue = Vref + Voe

      current = fabs((U_Value / 76.67)); // I in Amper (in absolut value)

      if((current > limit_current) && (relayState == 1))
          relayDisconnect(5);
  }

/*!
 * brief Measures the voltage offset of the current sensor.
 *
 * Measures the voltage offset of the current sensor. The value of the voltage offset is measured when the relay is disconnected, when the sensor measures zero current.
 */
  void currentCorrection(void)
  {
      adc16ChannelConfigStruct.channelNumber = DEMO_ADC16_USER_CHANNEL4;      //ADC4 (current (5))

      ADC16_SetChannelConfig(DEMO_ADC16_BASE, DEMO_ADC16_CHANNEL_GROUP, &adc16ChannelConfigStruct);
      while (0U == (kADC16_ChannelConversionDoneFlag & ADC16_GetChannelStatusFlags(DEMO_ADC16_BASE, DEMO_ADC16_CHANNEL_GROUP))) {}

      ADC_Value = ADC16_GetChannelConversionValue(DEMO_ADC16_BASE, DEMO_ADC16_CHANNEL_GROUP);   //Value measured from ADC0
      correctionValue = ((ADC_Value/ADC_range) * VCC_mV);   // U_Value from ADC0 in mV
  }

/*!
 * brief Calls functions of temperature measurement.
 *
 * Calls functions of temperature measurement with appropriate ADC channel. After measurement checks if measured value exceeded limit value. If yes, then relayDisconnect() is called with appropriate parameter.
 */
  void measureData(void)
  {
      adc16ChannelConfigStruct.channelNumber = DEMO_ADC16_USER_CHANNEL0;      //ADC0 (temperature 1)
      GPIO_PortSet(BOARD_INITPINS_TEMP1_GPIO, 1u << BOARD_INITPINS_TEMP1_PIN);
      temperature1 = temperatureMeasurement(&adc16ChannelConfigStruct);
      GPIO_PortClear(BOARD_INITPINS_TEMP1_GPIO, 1u << BOARD_INITPINS_TEMP1_PIN);
        if((temperature1 > limit_temperature1) && (relayState == 1))
          relayDisconnect(1);

      adc16ChannelConfigStruct.channelNumber = DEMO_ADC16_USER_CHANNEL1;      //ADC1 (temperature 2)
      GPIO_PortSet(BOARD_INITPINS_TEMP2_GPIO, 1u << BOARD_INITPINS_TEMP2_PIN);
      temperature2 = temperatureMeasurement(&adc16ChannelConfigStruct);
      GPIO_PortClear(BOARD_INITPINS_TEMP2_GPIO, 1u << BOARD_INITPINS_TEMP2_PIN);
        if((temperature2 > limit_temperature2) && (relayState == 1))
          relayDisconnect(2);

      adc16ChannelConfigStruct.channelNumber = DEMO_ADC16_USER_CHANNEL2;      //ADC2 (temperature 3)
      GPIO_PortSet(BOARD_INITPINS_TEMP3_GPIO, 1u << BOARD_INITPINS_TEMP3_PIN);
      temperature3 = temperatureMeasurement(&adc16ChannelConfigStruct);
      GPIO_PortClear(BOARD_INITPINS_TEMP3_GPIO, 1u << BOARD_INITPINS_TEMP3_PIN);
        if((temperature3 > limit_temperature3) && (relayState == 1))
          relayDisconnect(3);

      adc16ChannelConfigStruct.channelNumber = DEMO_ADC16_USER_CHANNEL3;      //ADC3 (temperature 4)
      GPIO_PortSet(BOARD_INITPINS_TEMP4_GPIO, 1u << BOARD_INITPINS_TEMP4_PIN);
      temperature4 = temperatureMeasurement(&adc16ChannelConfigStruct);
      GPIO_PortClear(BOARD_INITPINS_TEMP4_GPIO, 1u << BOARD_INITPINS_TEMP4_PIN);
        if((temperature4 > limit_temperature4) && (relayState == 1))
          relayDisconnect(4);
  }

/*!
 * brief Sets the buzzer.
 *
 * param state Specifies whether the buzzer should start or stop.
 */
  void setBuzzer(bool state)
  {
    if(state)
    {
      buzzer = true;
      GPIO_PortSet(BOARD_INITPINS_BUZZER_GPIO, 1u << BOARD_INITPINS_BUZZER_PIN);   //BUZZER ON
    }
    else
    {
      buzzer = false;
      GPIO_PortClear(BOARD_INITPINS_BUZZER_GPIO, 1u << BOARD_INITPINS_BUZZER_PIN); //BUZZER OFF
    }
  }

/*!
 * brief TPM interrupt handler.
 *
 * It counts interrupts and calls functions or sets variables according to the required time.
 *
 */
  void BOARD_TPM_HANDLER(void)
  {
      /* Clear interrupt flag.*/
      TPM_ClearStatusFlags(BOARD_TPM, kTPM_TimeOverflowFlag);

      currentMeasurement(); //measure current every ~ 30 ms

      ++interrupts;
      if(interrupts > 30)
      {
        measureData(); //measure temperature every ~ 1s
        interrupts = 0;
      }

      if(buzzer)
        {
          ++buzzerTimer;
        }

      if(buzzerTimer > 6)
      {
        setBuzzer(false);
        buzzerTimer = 0;
      }

      if(interruptDisabled)
      {
        ++interruptTimer;
      }

      if(interruptTimer > 30)
      {
        interruptDisabled = false;
        interruptTimer = 0;
      }
  }

/*!
 * brief PORTA (buttno) interrupt handler.
 *
 * Toggles state of relay if limit values are not exceeded and this interrupt is not temporarily disabled.
 */
  void PORTA_IRQ_HANDLER(void)
  {
    /* Clear interrupt flag.*/
    GPIO_PortClearInterruptFlags(BOARD_INITPINS_BUTTON_GPIO, 0xFFFF);

    if((relayState == 1) && (!interruptDisabled))
    {
      relayDisconnect(6);
    }
    else
    {
      if((temperature1 < limit_temperature1) && (temperature2 < limit_temperature2) && (temperature3 < limit_temperature3) && (temperature4 < limit_temperature4) && (current < limit_current) && (!interruptDisabled))
      {
        relayConnect();
      }
    }

    interruptDisabled = true;
  }

  /*******************************************************************************
   * TSA Table
   * Specifies which variables are accessible via FreeMASTER.
   ******************************************************************************/

  FMSTR_TSA_TABLE_BEGIN(TSAtable)

    FMSTR_TSA_RO_VAR(temperature1, FMSTR_TSA_DOUBLE)
    FMSTR_TSA_RO_VAR(temperature2, FMSTR_TSA_DOUBLE)
    FMSTR_TSA_RO_VAR(temperature3, FMSTR_TSA_DOUBLE)
    FMSTR_TSA_RO_VAR(temperature4, FMSTR_TSA_DOUBLE)
    FMSTR_TSA_RO_VAR(current, FMSTR_TSA_DOUBLE)

    FMSTR_TSA_RW_VAR(limit_temperature1, FMSTR_TSA_DOUBLE)
    FMSTR_TSA_RW_VAR(limit_temperature2, FMSTR_TSA_DOUBLE)
    FMSTR_TSA_RW_VAR(limit_temperature3, FMSTR_TSA_DOUBLE)
    FMSTR_TSA_RW_VAR(limit_temperature4, FMSTR_TSA_DOUBLE)
    FMSTR_TSA_RW_VAR(limit_current, FMSTR_TSA_DOUBLE)

    FMSTR_TSA_RO_VAR(disconnectionSrc, FMSTR_TSA_UINT16)
    FMSTR_TSA_RW_VAR(relayState, FMSTR_TSA_UINT16)
    FMSTR_TSA_RW_VAR(resetState, FMSTR_TSA_UINT16)
    FMSTR_TSA_RW_VAR(testValue, FMSTR_TSA_UINT16)
    FMSTR_TSA_RO_VAR(applicationID, FMSTR_TSA_UINT16)

  FMSTR_TSA_TABLE_END()


  FMSTR_TSA_TABLE_LIST_BEGIN()
    FMSTR_TSA_TABLE(TSAtable)
  FMSTR_TSA_TABLE_LIST_END()

  /*******************************************************************************
   * EOF
   ******************************************************************************/
