const FMNG_ConnectionBarStatus = {
  idle: 0,
  connecting: 1,
  connected: 2,
  notconnected: 3
};

(function() {
  if (window.jQuery === undefined) {
    console.error("FMNG_ConnectionBar.js requires jQuery");
    return;
  }

  var $ = window.jQuery;
  var _connMgr;

  /**
   * @class
   * @classdesc Connection Bar widget renders a bar in a host element, that has a connection button and connection status area. It can be extended with custom buttons.
   * To use Connection Bar in an application, folowing widget needs to be created first:
   * - {@link FMNG_ConnectionMgr Connection Manager}
   * @description Creates an instance of FMNG_ConnectionBar.
   * @this {FMNG_ConnectionBar}
   * @param {jQuery}             element The host element.
   * @param {FMNG_ConnectionMgr} connMgr Connection manager.
   *
   * @example
   * $(document).ready(function() {
   *     pcm = new PCM(window.location.host, onSocketOpen);
   * });
   * function (onSocketOpen) {
   *     // Initialize connection manager widget
   *     connectionMgr = new FMNG_ConnectionMgr(pcm);
   *     connectionBar = new FMNG_ConnectionBar($("#fmng_connection_bar_container"), connectionMgr);
   *     connectionBar.beforeConnect = beforeConnect;
   *     connectionBar.afterConnect = afterConnect;
   *     connectionBar.addButton("glyphicon-star").on("click", handleCustomBtnClick);
   * };
   *
   * function beforeConnect() {
   *     // set connection string before connecting
   *     connectionBar.setConnectionString("your connection string");
   * };
   *
   * function afterConnect(success) {
   *     if (success) {
   *         console.log("Target is now connected.");
   *     } else {
   *         console.log("Could not connect to the target.");
   *     }
   * };
   *
   * function handleCustomBtnClick() {
   *     alert("Custom button clicked");
   * };
   */
  this.FMNG_ConnectionBar = function(element, connMgr) {
    _connMgr = connMgr;
    this.beforeConnect = $.noop;
    this.afterConnect = $.noop;
    this.beforeDisconnect = $.noop;
    this.afterDisconnect = $.noop;
    this.status = FMNG_ConnectionBarStatus.idle;
    if (element) {
      this.render(element);
      this.registerEventHandlers();
    }
  };

  // Private methods

  /**
   * Toggles connection according to current status.
   * Triggers connection status change event with connecting status.
   * @ignore
   * @async
   * @this {FMNG_ConnectionBar}
   */
  async function toggleConnect() {
    let status = this.status;
    this.onConnectionStatusChange(FMNG_ConnectionBarStatus.connecting);
    switch (status) {
      case FMNG_ConnectionBarStatus.idle:
      case FMNG_ConnectionBarStatus.notconnected:
        await this.beforeConnect();
        status = await this.connect();
        await this.afterConnect(status);
        break;
      case FMNG_ConnectionBarStatus.connected:
        await this.beforeDisconnect();
        await this.disconnect();
        await this.afterDisconnect();
        break;
      default:
        return;
    }
  }

  // Public methods

  /**
   * On board connection status change event handler.
   *
   * @this {FMNG_ConnectionBar}
   * @param {number} status New status to transition to.
   */
  FMNG_ConnectionBar.prototype.onConnectionStatusChange = function(status) {
    if (this.status === status) {
      return;
    }
    if (this.host) {
      this.update(status);
    }
    this.status = status;
  };

  /**
   * Creates FMNG_ConnectionBar UI inside the provided host element.
   *
   * @this {FMNG_ConnectionBar}
   * @param {jQuery} host The host element.
   */
  FMNG_ConnectionBar.prototype.render = function(host) {
    this.host = host;
    this.connectionContainer = $("<div>").addClass("fmng_connection");
    this.connectButton = $("<a>").addClass("fmng_connection_icon connect glyphicon glyphicon-off");
    this.connectLabel = $("<label>").addClass("fmng_connection_message").text("Connect To Target");
    this.buttons = $("<div>").addClass("pull-right");
    this.connectionContainer
      .append(this.connectButton)
      .append(this.connectLabel)
      .append(this.buttons);
    this.host.append(this.connectionContainer);
  };

  /**
   * Registers connection widget UI event handlers.
   *
   * @this {FMNG_ConnectionBar}
   */
  FMNG_ConnectionBar.prototype.registerEventHandlers = function() {
    this.connectButton.on("click", toggleConnect.bind(this));
  };

  /**
   * Add new button to connection bar.
   *
   * @this {FMNG_ConnectionBar}
   * @return {jQuery} New created element
   */
  FMNG_ConnectionBar.prototype.addButton = function(glyphicon) {
    let button = $("<a>").addClass("fmng_connection_icon settings glyphicon " + glyphicon);
    this.buttons.append(button);
    return button;
  };

  /**
   * Updates UI according to connection status.
   *
   * @this {FMNG_ConnectionBar}
   * @param {jQuery} element The host element.
   */
  FMNG_ConnectionBar.prototype.update = function(status) {
    switch (status) {
      case FMNG_ConnectionBarStatus.idle:
        this.connectLabel.text("Target Disconnected: Connect To Target");
        this.connectButton.removeClass("connected glyphicon-transfer").addClass("glyphicon-off");
        break;
      case FMNG_ConnectionBarStatus.connecting:
        this.connectLabel.text("Connecting");
        this.connectButton.removeClass("glyphicon-off").addClass("connected glyphicon-transfer");
        break;
      case FMNG_ConnectionBarStatus.connected:
        this.connectLabel.text("Target Connected: " + this.connStr);
        break;
      case FMNG_ConnectionBarStatus.notconnected:
        this.connectLabel.text("Target Connection Failed: Check Settings & Try Again");
        this.connectButton.removeClass("connected");
        break;
    }
  };

  /**
   * Sets connection string.
   *
   * @this {FMNG_ConnectionBar}
   * @param {string} connStr Connection string.
   */
  FMNG_ConnectionBar.prototype.setConnectionString = function(connStr) {
    this.connStr = connStr;
  };

  /**
   * Opens communication channel.
   * Triggers connection status change event with:
   * - *connected* status on success
   * - *notconnected* on failure
   *
   * @async
   * @this {FMNG_ConnectionBar}
   * @return {boolean} *true* if connection succeeded, *false* otherwise.
   */
  FMNG_ConnectionBar.prototype.connect = async function() {
    if (await _connMgr.startComm(this.connStr) && await _connMgr.isBoardDetected()) {
      this.onConnectionStatusChange(FMNG_ConnectionBarStatus.connected);
      return true;
    }
    this.onConnectionStatusChange(FMNG_ConnectionBarStatus.notconnected);
    return false;
  };

  /**
   * Closes communication channel.
   * Triggers connection status change event with *idle* status.
   *
   * @async
   * @this {FMNG_ConnectionBar}
   * @return {boolean} *true* if connection succeeded.
   */
  FMNG_ConnectionBar.prototype.disconnect = async function() {
    _connMgr.stopComm();
    this.onConnectionStatusChange(FMNG_ConnectionBarStatus.idle);
  };

})();
