(function() {

  /**
   * Private variables
   */
  var _pcm;

  /**
   * @class
   * @classdesc Connection Manager widget manages connection control.
   * @description Creates an instance of FMNG_ConnectionMgr.
   * @param {PCM} pcm PCM class instance used for communicating with the target.
   */
  this.FMNG_ConnectionMgr = function(pcm) {
    _pcm = pcm;
  };

  /**
   * Opens communication port.
   * @see {@link PCM#StartComm StartComm}
   *
   * @async
   * @param {string} connStr Connection friendly name or connection string.
   * @return {boolean} *true* on success, *false* on failure.
   */
  FMNG_ConnectionMgr.prototype.startComm = async function(connStr) {
    try {
      await _pcm.StartComm(connStr);
      return true;
    } catch (err) {
      console.error(err);
      return false;
    }
  };

  /**
   * Closes currently open communication port.
   * @see {@link PCM#StopComm StopComm}
   *
   * @async
   * @return {boolean} *true* on success, *false* on failure.
   */
  FMNG_ConnectionMgr.prototype.stopComm = async function() {
    try {
      await _pcm.StopComm();
      return true;
    } catch (err) {
      console.error(err);
      return false;
    }
  };

  /**
   * Checks if the board is detected.
   * @see {@link PCM#IsBoardDetected IsBoardDetected}
   *
   * @async
   * @return {boolean} *true* if target is detected, *false* otherwise.
   */
  FMNG_ConnectionMgr.prototype.isBoardDetected = async function() {
    try {
      let response = await _pcm.IsBoardDetected();
      return response.data;
    } catch (err) {
      console.error(err);
      return false;
    }
  };

  /**
   * Returns list of named connection.
   * @see {@link PCM#EnumCommPorts EnumCommPorts}
   *
   * @async
   * @return {Array<string>} List of connection names.
   */
  FMNG_ConnectionMgr.prototype.getCommPorts = async function() {
    let ports = [];
    for (let i = 0; ; ++i) {
      try {
        let response = await _pcm.EnumCommPorts(i);
        ports.push(response.data);
      } catch (err) {
        break;
      }
    }
    return ports;
  };

  /**
   * Returns named connection info.
   * @see {@link PCM#GetCommPortInfo GetCommPortInfo}
   *
   * @async
   * @return {CommPortInfo} Communication port object.
   */
  FMNG_ConnectionMgr.prototype.getCommPortInfo = async function(name) {
    try {
      let response = await _pcm.GetCommPortInfo(name);
      return response.data;
    } catch (err) {
      return undefined;
    }
  };

  })();
  