(function() {

  /**
   * Private variables
   */
  var _pcm, _symbols;

  /**
   * Gets all the symbols from target as a map.
   * @ignore
   * @async
   */
  async function readSymbols() {
    for (let name in _symbols) {
      delete _symbols[name];
    }
    for (let i = 0; ; ++i) {
      try {
        let response = await _pcm.EnumSymbols(i);
        let name = response.data;
        response = await _pcm.GetSymbolInfo(name);
        _symbols[name] = {
          "addr": response.data.addr,
          "size": response.data.size,
          "type": response.data.type
        };
      } catch (err) {
        break;
      }
    }
  }

  /**
   * @class
   * @classdesc Symbol Manager widget handles embedded application symbol access.
   * @description Creates an instance of FMNG_SymbolMgr.
   * @param {PCM} pcm PCM class instance used for communicating with the target.
   */
  this.FMNG_SymbolMgr = function(pcm) {
      _pcm = pcm;
      _symbols = {};
  };

  /**
   * Reads TSA (Target Side Addressing) tables.
   * @see {@link PCM#ReadTSA ReadTSA}
   * @see {@link PCM#EnumSymbols EnumSymbols}
   * @see {@link PCM#GetSymbolInfo GetSymbolInfo}
   *
   * @async
   * @return {boolean} True if read succeeded.
   */
  FMNG_SymbolMgr.prototype.readTSA = async function() {
    try {
      let response = await _pcm.ReadTSA();
      if (response.data.count > 0) {
        await readSymbols();
      }
      return true;
    } catch (err) {
      return false;
    }
  };

  /**
   * Reads and parses elf file.
   * @see {@link PCM#ReadELF ReadELF}
   * @see {@link PCM#EnumSymbols EnumSymbols}
   * @see {@link PCM#GetSymbolInfo GetSymbolInfo}
   *
   * @async
   * @return {boolean} True if read succeeded.
   */
  FMNG_SymbolMgr.prototype.readELF = async function(elfPath) {
    try {
      let response = await _pcm.ReadELF(elfPath);
      if (response.data.count > 0) {
        await readSymbols();
      }
      return true;
    } catch (err) {
      return false;
    }
  };

  /**
   * @return {Object} Returns read symbols.
   */
  FMNG_SymbolMgr.prototype.getSymbols = function() {
    return _symbols;
  };

})();
